@@ -119,7 +119,6 @@
 ;;; Code:
 
 (require 'dig)
-(require 'comint)
 (eval-when-compile (require 'cl))
 
 (defgroup smime nil
@@ -194,6 +193,7 @@ If nil, use system defaults."
 
 (defvar smime-details-buffer "*OpenSSL output*")
 
+;; Use mm-util?
 (eval-and-compile
   (defalias 'smime-make-temp-file
     (if (fboundp 'make-temp-file)
@@ -210,8 +210,8 @@ If nil, use system defaults."
 (defun smime-ask-passphrase ()
   "Asks the passphrase to unlock the secret key."
   (let ((passphrase
-	 (comint-read-noecho
-	  "Passphrase for secret key (RET for no passphrase): " t)))
+	 (read-passwd
+	  "Passphrase for secret key (RET for no passphrase): ")))
     (if (string= passphrase "")
 	nil
       passphrase)))
@@ -489,6 +489,7 @@ A string or a list of strings is returned."
 	    (caddr curkey)
 	  (smime-get-certfiles keyfile otherkeys)))))
 
+;; Use mm-util?
 (eval-and-compile
   (defalias 'smime-point-at-eol
     (if (fboundp 'point-at-eol)
